// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

use std::{error, fmt, io, str};

use libudev;
use nix;
use uuid;
use serde_json;

use devicemapper;

#[derive(Debug, Clone)]
pub enum ErrorEnum {
    Error,

    AlreadyExists,
    Busy,
    Invalid,
    NotFound,
}

#[derive(Debug)]
pub enum EngineError {
    Engine(ErrorEnum, String),
    Io(io::Error),
    Nix(nix::Error),
    Uuid(uuid::ParseError),
    Utf8(str::Utf8Error),
    Serde(serde_json::error::Error),
    DM(devicemapper::DmError),
    Udev(libudev::Error),
}

impl fmt::Display for EngineError {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match *self {
            EngineError::Engine(_, ref msg) => write!(f, "Stratis error: {}", msg),
            EngineError::Io(ref err) => write!(f, "IO error: {}", err),
            EngineError::Nix(ref err) => write!(f, "Nix error: {}", err),
            EngineError::Uuid(ref err) => write!(f, "Uuid error: {}", err),
            EngineError::Utf8(ref err) => write!(f, "Utf8 error: {}", err),
            EngineError::Serde(ref err) => write!(f, "Serde error: {}", err),
            EngineError::DM(ref err) => write!(f, "DM error: {}", err),
            EngineError::Udev(ref err) => write!(f, "udev error: {}", err),
        }
    }
}

impl error::Error for EngineError {
    fn description(&self) -> &str {
        match *self {
            EngineError::Engine(_, ref msg) => msg,
            EngineError::Io(ref err) => err.description(),
            EngineError::Nix(ref err) => err.description(),
            EngineError::Uuid(_) => "Uuid::ParseError",
            EngineError::Utf8(ref err) => err.description(),
            EngineError::Serde(ref err) => err.description(),
            EngineError::DM(ref err) => err.description(),
            EngineError::Udev(ref err) => err.description(),
        }
    }
}

pub type EngineResult<T> = Result<T, EngineError>;

impl From<io::Error> for EngineError {
    fn from(err: io::Error) -> EngineError {
        EngineError::Io(err)
    }
}

impl From<nix::Error> for EngineError {
    fn from(err: nix::Error) -> EngineError {
        EngineError::Nix(err)
    }
}

impl From<uuid::ParseError> for EngineError {
    fn from(err: uuid::ParseError) -> EngineError {
        EngineError::Uuid(err)
    }
}

impl From<str::Utf8Error> for EngineError {
    fn from(err: str::Utf8Error) -> EngineError {
        EngineError::Utf8(err)
    }
}

impl From<serde_json::error::Error> for EngineError {
    fn from(err: serde_json::error::Error) -> EngineError {
        EngineError::Serde(err)
    }
}

impl From<devicemapper::DmError> for EngineError {
    fn from(err: devicemapper::DmError) -> EngineError {
        EngineError::DM(err)
    }
}

impl From<libudev::Error> for EngineError {
    fn from(err: libudev::Error) -> EngineError {
        EngineError::Udev(err)
    }
}
